!#################################################################
!
! contains calls to user-specific output routines
!
!### macros #####################################################
!
#define TRACEBACK write (gol,'("in ",a," (",a,", line",i5,")")') rname, __FILE__, __LINE__; call goErr
#define IF_NOTOK_RETURN(action) if (status/=0) then; TRACEBACK; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; TRACEBACK; action; return; end if
!
#include "tm5.inc"
!
!-----------------------------------------------------------------------------
!                    TM5                                                     !
!-----------------------------------------------------------------------------
!BOP
!
! !MODULE:      USER_OUTPUT
!
! !DESCRIPTION:  Contains calls to user-specific output routines, e.g.
!                instantaneous mix files, station output, output of flight 
!                tracks etc.
!\\
!\\
! !INTERFACE: 
!
MODULE USER_OUTPUT

  use GO, only : gol, goPr, goErr, goBug, goLabel

  implicit none

  ! --- in/out ------------------------

  private

  public                      :: user_output_step, user_output_init, user_output_done, user_output_mean
  public                      :: column_data_dtsec

  ! --- const -------------------------

  character(len=*), parameter  ::  mname = 'user_output'
  
  ! --- var ---------------------------

  logical                     :: settings_data = .false.   ! signal for model settings output

  logical                     :: mix_data = .false.        ! signal for mix output
  integer                     :: mix_data_dhour            ! every dhour hour

  logical                     :: column_data = .false.     ! signal for column output
  integer                     :: column_data_dtsec         ! every dhour hour

  logical                     :: mmix_data = .false.     ! signal for mean mix output

  logical                     :: flask_data = .false.  ! signal for flask output

!  logical           :: station_data = .true.  ! signal for station output

!  logical           :: column3d_data = .false.  ! signal for columnoutput
!  integer           :: column3d_data_dhour=24         ! every dhour hour

!  logical           :: flux1x1_data = .false.     ! signal for flux1x1 output
!  integer           :: flux1x1_data_dhour         ! every dhour hour

#ifdef with_retro_output
  logical           ::  output_retro = .false.
  integer           ::  output_retro_dhour         ! every dhour hour
#endif


#ifdef with_cf_output
  logical           ::  output_cf = .false.
  integer           ::  output_cf_dhour         ! every dhour hour
#endif

contains


  !-------------------------------------------------------------
  ! user_output_init:
  !      Initialise user-specified model output (all regions)
  !-------------------------------------------------------------

  subroutine user_output_init( status )

    use dims,                   only: itau, ndyn_max
    use GO,                     only : TrcFile, Init, Done, ReadRc
    use global_data,            only : rcfile
    use user_output_settings,   only : user_output_settings_init
    use user_output_mix,        only : user_output_mix_init
    use user_output_column,     only : user_output_column_init
 !   use user_output_flux1x1, only : init_flux1x1
 !   use user_output_station, only : read_stationlist, init_station_output, StationlistFFilename 
    use user_output_mmix,       only : mmix_init
    use user_output_flask,      only : user_output_flask_init
    use user_output_common,	only : user_output_common_init
#ifdef with_retro_output
    use user_output_retro,      only : output_retro_init
#endif
#ifdef with_cf_output
    use user_output_cf,         only : output_cf_init
#endif
 
    ! --- in/out -----------------------------
    
    integer, intent(inout)      ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/user_output_init'
    
    ! --- local -------------------------------

    type(TrcFile)               ::  rcF

    ! --- begin -------------------------------
    
    ! init common stuff:
    call User_Output_Common_Init( status )
    IF_NOTOK_RETURN(status=1)

    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)

    ! output settings ?
    call ReadRc( rcF, 'settings.output', settings_data, status, default=.false. )
    IF_ERROR_RETURN(status=1)

    ! apply ?
    if ( settings_data ) then
      write (gol,'(a,":   user output settings ...")') rname; call goPr
      call user_output_settings_init( rcF, status )
      IF_NOTOK_RETURN(status=1)
    end if 

!    call ReadRc( rcF, 'output.station', station_data, status )
!    IF_NOTOK_RETURN(status=1)
!    if ( station_data ) then
!       call read_stationlist(status)
!       IF_NOTOK_RETURN(status=1)
!       call init_station_output(status)
!       IF_NOTOK_RETURN(status=1)
!    end if

    call ReadRc( rcF, 'output.mix', mix_data, status )
    IF_NOTOK_RETURN(status=1)
    if ( mix_data ) then
      call ReadRc( rcF, 'output.mix.dhour', mix_data_dhour, status )
      write (gol,*) trim(mname)//'/mix_data_dhour:', mix_data_dhour; call goPr
      IF_NOTOK_RETURN(status=1)
      call user_output_mix_init(status)
      IF_NOTOK_RETURN(status=1)
    end if

    ! initialise accumulation of the mean mixing ratio fields
    call ReadRc( rcF, 'output.mmix', mmix_data, status )
    IF_NOTOK_RETURN(status=1)
    if ( mmix_data ) then
       call mmix_init(status)
       IF_NOTOK_RETURN(status=1)
    end if  

    ! Note that the flux1x1 module uses the emissions_averagers code.  That
    ! code was originally written to accumulate weekly average posterior fluxes
    ! during an inverse run, and it is handled with calls scattered throughout the
    ! EnKF modules.  If an inverse run also specifies, output.flux1x1:T in its
    ! rc file, then the weekly average emissions (which, incidentally get written
    ! to the savestate.hdf file, not flux1x1*hdf files) need to be turned off.
    !  - Andy, 15 Nov 08
!    call ReadRc( rcF, 'output.flux1x1', flux1x1_data, status ,default=.false.)
!    IF_NOTOK_RETURN(status=1)
!    if ( flux1x1_data ) then
!      call ReadRc( rcF, 'output.flux1x1.dhour', flux1x1_data_dhour, status , default=24)
!      IF_NOTOK_RETURN(status=1)
!      call init_flux1x1
!    end if

    call ReadRc( rcF, 'output.flask', flask_data, status, default=.false.)
    IF_NOTOK_RETURN(status=1)
    if(flask_data) then
       call user_output_flask_init(rcF,status)
       IF_NOTOK_RETURN(status=1)
    endif
    
#ifdef with_retro_output
    ! put out in retro format ?
    call ReadRc( rcF, 'output.retro', output_retro, status, default=.false. )
    IF_ERROR_RETURN(status=1)
    ! init if necessary:
    if ( output_retro ) then
      ! init output; return ouptut time step
      call Output_RETRO_Init( rcF, output_retro_dhour, status )
      IF_NOTOK_RETURN(status=1)
    end if
#endif

#ifdef with_cf_output
    ! put out in retro format ?
    call ReadRc( rcF, 'output.cf', output_cf, status, default=.false. )
    IF_ERROR_RETURN(status=1)
    ! init if necessary:
    if ( output_cf ) then
      ! init output; return ouptut time step
      call Output_CF_Init( rcF, output_cf_dhour, status )
      IF_NOTOK_RETURN(status=1)
    end if
#endif

    ! output column data?
    call ReadRc( rcF, 'output.column', column_data, status, default = .false. )
    IF_NOTOK_RETURN( status=1 )

    ! apply?
    if ( column_data ) then
      call ReadRc( rcF, 'output.column.dtsec', column_data_dtsec, status , default = ndyn_max)
      IF_ERROR_RETURN(status=1)
      write (gol,*) trim(mname)//'/column_data_dtsec:', column_data_dtsec; call goPr
      call user_output_column_init( status )
      IF_NOTOK_RETURN( status=1 )
    end if

    ! close rcfile:
    call Done( rcF, status  )
    IF_NOTOK_RETURN(status=1)

    ! ok
    status = 0

  end subroutine user_output_init


  !-------------------------------------------------------------
  ! user_output_done:
  !      Finalise user-specified model output for the region given
  !-------------------------------------------------------------

  subroutine user_output_done( status )

    use user_output_settings,   only : user_output_settings_done
    use user_output_mix,        only : user_output_mix_done
    use user_output_column,     only : user_output_column_done 
    use dims,                only : nregions
!    use user_output_station,  only : close_stationfiles, write_stationfiles
    use user_output_mmix,    only : write_mmix, mmix_Done
 !   use user_output_column3d,   only : free_columndata3d  
    use user_output_flask,	only : user_output_flask_done
    use User_Output_Common,	only : User_Output_Common_Done

#ifdef with_retro_output
    use user_output_retro  , only : output_retro_done
#endif
#ifdef with_cf_output
    use user_output_cf     , only : output_cf_done
#endif
    
    ! --- in/out -----------------------------
    
    integer, intent(out)      ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/user_output_done'
    
    ! --- begin -----------------------------

    ! settings output enabled ?
    if ( settings_data ) then
      ! done with module:
      call User_Output_Settings_Done( status )
      IF_NOTOK_RETURN(status=1)
    end if 
    
    ! write the mean mixing ratio fields to file
    if ( mmix_data ) then
      call write_mmix(status)
      IF_NOTOK_RETURN(status=1)
      call mmix_Done(status)
      IF_NOTOK_RETURN(status=1)      
    end if
    
!    if(column3d_data ) call free_columndata3d 
    
!    if ( write_stationfiles .and. station_data) call close_stationfiles

    if(flask_data) then
       call user_output_flask_done(status)
      IF_NOTOK_RETURN(status=1)
    endif

    if ( mix_data ) then 
       call user_output_mix_done( status )
       IF_NOTOK_RETURN(status=1)
    endif
    
#ifdef with_retro_output
    if ( output_retro ) then
      call Output_Retro_Done( status )
      IF_NOTOK_RETURN(status=1)
    end if
#endif

#ifdef with_cf_output
    if ( output_cf ) then
      call Output_cf_Done( status )
      IF_NOTOK_RETURN(status=1)
    end if
#endif
    
    if ( column_data ) then
      call user_output_column_done( status )
      IF_NOTOK_RETURN( status=1 )
    endif

    ! done with common stuff:
    call User_Output_Common_Done( status )
    IF_NOTOK_RETURN(status=1)
    
    ! ok
    status = 0

  end subroutine user_output_done

  !-------------------------------------------------------------
  ! user_output_step:
  !      Define user-specified model output for the region given
  !      Called every time step
  !-------------------------------------------------------------

  subroutine user_output_step( region, status )

    use dims,                 only : itaur, itaui, itau
    use datetime,             only : tau2date
!    use user_output_station,  only : output_stationconc  
    use user_output_mix,          only : user_output_mix_accum, user_output_mix_write
    use user_output_column,       only: user_output_column_accum,user_output_column_evaluate,user_output_column_write,user_output_column_reset
    use user_output_mmix,     only : accumulate_mmix
    use user_output_flask,		only : user_output_flask_sample
#ifdef with_retro_output
    use user_output_retro   , only : Output_Retro_Step
#endif
#ifdef with_cf_output
    use user_output_cf      , only : Output_cf_Step
#endif

    ! --- in/out ------------------------------
    
    integer, intent(in)       ::  region
    integer, intent(out)      ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/user_output_step'
    
    ! --- local ------------------------------
    
    integer,dimension(6)            :: idate_f

    ! --- begin ------------------------------
    
    call tau2date(itaur(region),idate_f)

    if ( mmix_data ) then
       call accumulate_mmix( region,status )
       IF_NOTOK_RETURN(status=1)
    endif
    

!    if ( station_data ) then
!       call output_stationconc(region, status)
!       IF_NOTOK_RETURN(status=1)
!    endif

    if ( mix_data ) then
       call user_output_mix_accum(region,status)
       if ( (itaur(region)-itaui .gt. 0) .and. (modulo(itaur(region)-itaui,mix_data_dhour*3600) == 0 )) then
          call user_output_mix_write(region, status )
          IF_NOTOK_RETURN(status=1)
       endif
       !call output_mix(region)
    end if

    if (column_data) then
      call user_output_column_accum( region, status )
      IF_NOTOK_RETURN( status=1 )
       if ( (itaur(region)-itaui .gt. 0) .and. (modulo(itaur(region)-itaui,column_data_dtsec) == 0 )) then
          call user_output_column_evaluate( status )
          IF_NOTOK_RETURN(status=1)
          call user_output_column_write( region, status )
          IF_NOTOK_RETURN(status=1)
          call user_output_column_reset( status )
          IF_NOTOK_RETURN(status=1)
       endif
    endif

    if(flask_data) then
       call user_output_flask_sample(region,itaur(region),status)
      IF_NOTOK_RETURN(status=1)
    endif

#ifdef with_retro_output
    if ( output_retro ) then
      if ( (modulo(idate_f(4),output_retro_dhour)==0) .and. all(idate_f(5:6)==0) ) then
        call Output_Retro_Step( region, idate_f, status )
        IF_NOTOK_RETURN(status=1)
      end if
    end if
#endif

#ifdef with_cf_output
    if ( output_cf ) then
      if ( (modulo(idate_f(4),output_cf_dhour)==0) .and. all(idate_f(5:6)==0) ) then
        call Output_CF_Step( region, idate_f, status )
        IF_NOTOK_RETURN(status=1)
      end if
    end if
#endif

    ! ok
    status = 0

  end subroutine user_output_step
  
  !-------------------------------------------------------------
  ! user_output_mean:
  !      called from tm5 main
  !-------------------------------------------------------------
  
  subroutine user_output_mean( status )

    use datetime,                 only : tau2date
    use dims,                     only : nregions, itau, ndyn_max
    use user_output_column,       only : user_output_column_write, user_output_column_evaluate, user_output_column_reset
!  use user_output_station, only : evaluate_stationconc, reset_stationconc_accumulator, write_stationconc

    implicit none

    ! --- in/out ------------------------------
    
    integer, intent(inout)    :: status
    integer                   :: region
    integer,dimension(6)      :: idate_f

    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/user_output_mean'

 !   IF(station_data)THEN
 !     IF(mod(itau, ndyn_max) == 0) THEN
 !        CALL evaluate_stationconc(status)
 !        IF_NOTOK_RETURN(status=1)
 !        CALL write_stationconc(status)
 !        IF_NOTOK_RETURN(status=1)
 !        CALL reset_stationconc_accumulator
 !     ENDIF
 !   ENDIF
  
    ! ok
    status = 0

  end subroutine user_output_mean
  
end module user_output
